"""Publish messages based on Minio bucket notifications.

The published messages will contain filesystem information generated by fsspec.
"""

from logging import getLogger

from upath import UPath

from pytroll_watchers.publisher import SecurityError, file_publisher_from_generator, parse_metadata

logger = getLogger(__name__)


def file_publisher(config):
    """Publish objects coming from bucket notifications.

    Args:
        config: the configuration dictionary, containing in particular an fs_config section, which is the configuration
        for the filesystem watching, will be passed as argument to `file_generator`. The other sections are passed
        further to ``file_publisher_from_generator``.
    """
    fs_config = config["fs_config"]
    logger.info(f"Starting watch on '{fs_config['bucket_name']}'")
    if "secret_key" in fs_config.get("storage_options", []):
        raise SecurityError("A secret key cannot be published safely.")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, config)


def file_generator(endpoint_url, bucket_name, file_pattern=None, storage_options=None):
    """Generate new objects appearing in the watched bucket.

    Args:
        endpoint_url: The endpoint_url to use.
        bucket_name: The bucket to watch for changes.
        file_pattern: The trollsift pattern to use for matching and extracting metadata from the object name.
            This can include the prefix if needed.
        storage_options: The storage options for the service, for example for specifying a profile to the aws config.

    Returns:
        A tuple of UPath and metadata.

    Examples:
        To iterate over new files in `s3:///tmp/`:

        >>> for filename in file_generator("some_endpoint_url", "tmp",
        ...                                file_pattern="data/{start_time:%Y%m%d_%H%M}_{product}.tif")
        ...    print(filename)
        UPath("s3:///tmp/data/20200428_1000_foo.tif")

    """
    object_metadata = {}

    if storage_options is None:
        storage_options = {}
    for record in _record_generator(endpoint_url, bucket_name, storage_options):
        for item in record["Records"]:
            new_bucket_name = item["s3"]["bucket"]["name"]
            object_name = item["s3"]["object"]["key"]
            try:
                object_metadata = parse_metadata(file_pattern, object_name)
            except ValueError:
                continue

            path = UPath(f"s3://{new_bucket_name}/{object_name}", **storage_options)
            yield path, object_metadata


def _record_generator(endpoint_url, bucket_name, storage_options):
    """Generate records for new objects in the bucket."""
    from minio import Minio
    from minio.credentials.providers import AWSConfigProvider

    if "profile" in storage_options:
        credentials = AWSConfigProvider(profile=storage_options["profile"])
    else:
        credentials = None

    client = Minio(endpoint_url,
        credentials=credentials
    )

    with client.listen_bucket_notification(
        bucket_name,
        # prefix="my-prefix/",
        events=["s3:ObjectCreated:*"],
    ) as events:
        for event in events:
            yield event



"""Collection of function needed by multiple watchers."""

import datetime
import time


def run_every(interval):
    """Generator that ticks every `interval`.

    Args:
        interval: the timedelta object giving the amount of time to wait between ticks. An interval of 0 will just make
        tick once, then return (and thus busy loops aren't allowed).

    Yields:
        The time of the next tick.
    """
    while True:
        next_check = datetime.datetime.now(datetime.timezone.utc) + interval
        yield next_check
        to_wait = max(next_check.timestamp() - time.time(), 0)
        time.sleep(to_wait)
        if not interval:  # interval is 0
           break


def fromisoformat(datestring):
    """Wrapper around datetime's fromisoformat that also works on python 3.10."""
    try:
        return datetime.datetime.fromisoformat(datestring)
    except ValueError:
        # for python 3.10
        return datetime.datetime.strptime(datestring, "%Y-%m-%dT%H:%M:%S.%f%z")



"""Fetch files from other (remote) filesystems."""

import json
import logging
import os
from pathlib import Path
from urllib.parse import unquote

import fsspec

logger = logging.getLogger(__name__)


def fetch_file(file_to_fetch, download_dir, filesystem=None):
    """Fetch a file.

    Args:
        file_to_fetch: The Path of the file to fetch. Can be a UPath from universal_path.
        download_dir: The directory to download the file to.
        filesystem: The file system to use if provided. Should be a dictionary that will be fed to fsspec.

    Returns:
        The Path to the downloaded file.

    Example:

        >>> fetch_file("https://noaa-himawari8.s3.amazonaws.com/AHI-L1b-FLDK/2017/02/02/0020/HS_H08_20170202_0020_B01_FLDK_R10_S0101.DAT.bz2", "/tmp/")
        ... # HS_H08_20170202_0020_B01_FLDK_R10_S0101.DAT.bz2 is now present in /tmp/


    """  # noqa
    download_dir = Path(download_dir)

    if filesystem:
        downloaded_file = _fetch_from_json_filesystem(file_to_fetch, download_dir, filesystem)
    else:
        downloaded_file = _fetch_from_uri(file_to_fetch, download_dir)
    logger.info(f"Fetched {str(downloaded_file)}")
    return downloaded_file


def _fetch_from_uri(file_to_fetch, download_dir):
    """Fetch a file from a uri."""
    fs_file = fsspec.open(file_to_fetch)
    filesystem = fs_file.fs
    return _fetch_from_filesystem(fs_file.path, download_dir, filesystem)


def _fetch_from_json_filesystem(path_to_fetch, download_dir, fs):
    """Fetch a file from a path and a filesystem specification."""
    filesystem = fsspec.AbstractFileSystem.from_json(json.dumps(fs))
    return _fetch_from_filesystem(path_to_fetch, download_dir, filesystem)


def _fetch_from_filesystem(path_to_fetch, download_dir, fs):
    """Fetch a file from a path and a filesystem."""
    basename = unquote(os.path.basename(path_to_fetch))
    downloaded_file = download_dir / basename
    return _fetch_file(path_to_fetch, downloaded_file, fs)


def _fetch_file(remote_path, local_path, fs):
    """Fetch the remote path to the local path."""
    fs.get_file(remote_path, local_path)
    return local_path



"""Main interface functions."""

import argparse
import logging.config
from importlib.metadata import entry_points

import yaml


def get_publisher_for_backend(backend):
    """Get the right publisher for the given backend.

    For the parameters to pass the returned function, check the individual backend documentation pages.

    Example:
        >>> file_publisher = get_publisher_for_backend("local")
        >>> file_publisher(fs_config, publisher_config, message_config)

    """
    return get_backend(backend).file_publisher


def get_backend(backend):
    """Get the module for a given backend."""
    eps = entry_points(group="pytroll_watchers.backends")
    for ep in eps:
        if backend == ep.name:
            return ep.load()
    raise ValueError(f"Unknown backend {backend}.")


def get_generator_for_backend(backend):
    """Get the right generator for the given backend.

    For the parameters to pass the returned function, check the individual backend documentation pages.

    Example:
        >>> file_generator = get_generator_for_backend("local")
        >>> for filename, file_metadata in file_generator("/tmp"):
        ...     # do something with filename and file_metadata

    """
    return get_backend(backend).file_generator


def publish_from_config(config):
    """Publish files/objects given a config.

    Args:
        config: a dictionary containing the `backend` string (`local` or `minio`), and `fs_config`, `publisher_config`
            and `message_config` dictionaries.
    """
    backend = config["backend"]
    publisher = get_publisher_for_backend(backend)
    return publisher(config)



def cli(args=None):
    """Command-line interface for pytroll-watchers."""
    parser = argparse.ArgumentParser(
                    prog="pytroll-watcher",
                    description="Watches the appearance of new files/objects on different filesystems.",
                    epilog="Thanks for using pytroll-watchers!")

    parser.add_argument("config", type=str, help="The yaml config file.")
    parser.add_argument("-l", "--log-config", type=str, help="The yaml config file for logging.", default=None)

    parsed = parser.parse_args(args)


    log_config_filename = parsed.log_config
    configure_logging(log_config_filename)

    config_file = parsed.config

    with open(config_file) as fd:
        config_dict = yaml.safe_load(fd.read())

    return publish_from_config(config_dict)


def configure_logging(log_config_filename):
    """Configure logging from a yaml file."""
    if log_config_filename is not None:
        with open(log_config_filename) as fd:
            log_config = yaml.safe_load(fd.read())
    else:
        log_config = {
            "version": 1,
            "formatters": {
                "pytroll": {
                    "format": "[%(asctime)s %(levelname)-8s %(name)s] %(message)s"
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "INFO",
                    "formatter": "pytroll",
                },
            },
            "disable_existing_loggers": False,
            "loggers": {
                "": {
                    "level": "INFO",
                    "handlers": ["console"],
                    "propagate": True
                },
            },
        }
    logging.config.dictConfig(log_config)



"""Test the local watcher."""

import os

import pytest
from posttroll.message import Message
from posttroll.testing import patched_publisher

from pytroll_watchers import local_watcher
from pytroll_watchers.publisher import SecurityError
from pytroll_watchers.testing import patched_local_events  # noqa


def test_watchdog_generator_with_os(tmp_path, patched_local_events):  # noqa
    """Test a watchdog generator."""
    filename = os.fspath(tmp_path / "20200428_1000_foo.tif")

    with patched_local_events([filename]):
        fname_pattern = "{start_time:%Y%m%d_%H%M}_{product}.tif"

        generator = local_watcher.file_generator(tmp_path,
                                                "os",
                                                file_pattern=fname_pattern)
    path, metadata = next(generator)

    assert str(path) == filename
    assert metadata["product"] == "foo"


def test_watchdog_generator_with_protocol(tmp_path, patched_local_events):  # noqa
    """Test a watchdog generator."""
    filename = os.fspath(tmp_path / "20200428_1000_foo.tif")

    with patched_local_events([filename]):
        fname_pattern = "{start_time:%Y%m%d_%H%M}_{product}.tif"

        protocol = "ssh"
        storage_options = {"parameter": "value",
                           "host": "somehost.pytroll.org"}


        generator = local_watcher.file_generator(tmp_path,
                                                 file_pattern=fname_pattern,
                                                 protocol=protocol,
                                                 storage_options=storage_options)
    path, metadata = next(generator)

    assert path.as_uri().startswith("ssh://")
    assert path.as_uri().endswith(filename)
    assert path.protocol == protocol
    assert path.storage_options == storage_options
    assert metadata["product"] == "foo"


def test_watchdog_generator_with_polling(tmp_path, patched_local_events):  # noqa
    """Test a watchdog generator."""
    filename = os.fspath(tmp_path / "20200428_1000_foo.tif")
    with patched_local_events([filename]):
        fname_pattern = "{start_time:%Y%m%d_%H%M}_{product}.tif"
        generator = local_watcher.file_generator(tmp_path,
                                                 "polling",
                                                 file_pattern=fname_pattern)

    path, _ = list(generator)[0]
    assert str(path) == filename


def test_watchdog_generator_with_something_else(tmp_path):
    """Test a watchdog generator."""
    fname_pattern = "{start_time:%Y%m%d_%H%M}_{product}.tif"

    generator = local_watcher.file_generator(tmp_path,
                                             "something_else",
                                             file_pattern=fname_pattern)

    with pytest.raises(ValueError, match="'os' or 'polling'"):
        next(generator)


def test_publish_paths(tmp_path, patched_local_events, caplog):  # noqa
    """Test publishing paths."""
    basename = "foo+bar,baz_.txt"
    filename = os.fspath(tmp_path / basename)

    local_settings = dict(directory=tmp_path)
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))

    caplog.set_level("INFO")
    with patched_local_events([filename]):
        with patched_publisher() as messages:
            local_watcher.file_publisher(dict(fs_config=local_settings,
                                              publisher_config=publisher_settings,
                                              message_config=message_settings))

    assert "uri" not in message_settings["data"]
    assert len(messages) == 1
    message = Message(rawstr=messages[0])
    assert message.data["uri"] == f"{str(tmp_path)}/{basename}"
    assert message.data["sensor"] == "viirs"
    assert "filesystem" not in message.data
    assert f"Starting watch on '{local_settings['directory']}'" in caplog.text


def test_publish_paths_forbids_passing_password(tmp_path, patched_local_events, caplog):  # noqa
    """Test publishing paths."""
    filename = os.fspath(tmp_path / "foo.txt")
    password = "very strong"  # noqa

    local_settings = dict(directory=tmp_path, protocol="ssh",
                          storage_options=dict(host="myhost.pytroll.org", username="user", password=password))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))

    caplog.set_level("INFO")
    with patched_local_events([filename]):
        with patched_publisher():
            with pytest.raises(SecurityError):
                local_watcher.file_publisher(dict(fs_config=local_settings,
                                                  publisher_config=publisher_settings,
                                                  message_config=message_settings))


def test_publish_paths_with_ssh(tmp_path, patched_local_events):  # noqa
    """Test publishing paths with an ssh protocol."""
    filename = os.fspath(tmp_path / "foo.txt")

    host = "localhost"

    local_settings = dict(directory=tmp_path, protocol="ssh",
                          storage_options=dict(host=host))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))

    with patched_local_events([filename]):
        with patched_publisher() as published_messages:
            local_watcher.file_publisher(dict(fs_config=local_settings,
                                              publisher_config=publisher_settings,
                                              message_config=message_settings))
            assert len(published_messages) == 1
            message = Message(rawstr=published_messages[0])
            assert message.data["uri"].startswith("ssh://")
            assert message.data["filesystem"]["host"] == host


def test_publish_paths_with_file(tmp_path, patched_local_events):  # noqa
    """Test publishing paths with a file protocol."""
    filename = os.fspath(tmp_path / "foo.txt")

    local_settings = dict(directory=tmp_path, protocol="file")
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))

    with patched_local_events([filename]):
        with patched_publisher() as published_messages:
            local_watcher.file_publisher(dict(fs_config=local_settings,
                                              publisher_config=publisher_settings,
                                              message_config=message_settings))
            assert len(published_messages) == 1
            message = Message(rawstr=published_messages[0])
            assert message.data["uri"].startswith("file://")
            assert "filesystem" in message.data



"""Common functions for publishing messages."""

import datetime
import json
import logging
from contextlib import closing, contextmanager, suppress
from copy import deepcopy
from urllib.parse import unquote
from warnings import warn

import fsspec
from posttroll.message import Message
from posttroll.publisher import create_publisher_from_dict_config
from trollsift import parse
from upath import UPath

from pytroll_watchers.fetch import fetch_file

logger = logging.getLogger(__name__)


class SecurityError(Exception):
    """An exception for breaking security rules."""


def file_publisher_from_generator(generator, config):
    """Publish files coming from local filesystem events.

    Args:
        generator: the generator to use for producing files. The generator must yield tuples of
            (filename, file_metadata).
        config: the configuration containing the parameters to use for publishing data. It should contain the following
                sections:
                  - publisher_config: The configuration dictionary to pass to the posttroll publishing functions.
                  - message_config: The information needed to complete the posttroll message generation. Will be amended
                      with the file metadata, and passed directly to posttroll's Message constructor.
                  - data_config: The information about the processing to do on the uris before sending.
                      The `fetch` section, if present, will trigger the fetching of the file locally before sendig the
                      (adjusted) uri further. The parameter `destination` should be provided as the directory to put
                      the downloaded files in.
                      The `unpack` section that contains the packing `format` for the archive (eg "zip"), or
                      "directory". The contents of the archive or directory will be published as a "dataset". For the case where
                      "directory" is used, it is also possible to set the boolean "include_dir_in_uid" to true so that
                      the full relative path of the file is provided (False by default).

    Side effect:
        Publishes posttroll messages containing the location of the file with the following fields:
          - The "uid" which is the unique filename of the file.
          - The "uri" which provides an fsspec-style uri to the file. This does not however contain the connection
            parameters, or storage options of the filesystems referred to.
          - The "filesystem" which is the json-serialized filesystem information that can then be fed to fsspec.
          - The "path" which is the path of the file inside the filesystem.

        For example, the file
          - uid: `S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3/Oa02_radiance.nc`
          - uri: `s3:///eodata/Sentinel-3/OLCI/OL_1_EFR___/2024/04/15/S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3/Oa02_radiance.nc`
          - filesystem: `{"cls": "s3fs.core.S3FileSystem", "protocol": "s3", "args": [], "profile": "my_profile"}`
          - path: `/eodata/Sentinel-3/OLCI/OL_1_EFR___/2024/04/15/S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3/Oa02_radiance.nc`
    """  # noqa
    publisher_config = config.pop("publisher_config")
    publisher = create_publisher_from_dict_config(publisher_config)
    publisher.start()


    with closing(publisher):
        for file_item, file_metadata in generator:
            msg = _create_message(file_item, file_metadata, config)
            logger.info(f"Sending {str(msg)}")
            publisher.send(str(msg))


def _create_message(file_item, file_metadata, config):
    config = deepcopy(config)
    message_config = config.pop("message_config", dict())
    unpack = message_config.pop("unpack", None)
    if unpack is not None:
        warn("The `unpack` option should be passed inside the `data_config` section", DeprecationWarning, stacklevel=1)

    data_config = config.pop("data_config", dict())

    if file_metadata and ("data" in file_metadata):
        file_mda = deepcopy(file_metadata)
        message_data = file_mda.pop("data")
        message_parameters = file_mda
    else:
        message_data = file_metadata or dict()
        message_parameters = dict()
    message_parameters.update(message_config)
    message_parameters.setdefault("data", {})

    file_location_info = prepare_data(file_item, data_config)
    message_parameters["data"].update(file_location_info)

    aliases = message_parameters.pop("aliases", {})
    apply_aliases(aliases, message_data)

    message_parameters["data"].update(message_data)

    return Message(**message_parameters)


def prepare_data(file_item, data_config):
    """Prepare the data for further processing."""
    fetch = data_config.pop("fetch", {})
    if fetch:
        file_item = fetch_file(file_item, fetch["destination"])

    unpack_info = data_config.pop("unpack", {})
    unpack = unpack_info.get("format", None)

    include_dir = unpack_info.get("include_dir_in_uid", False)

    metadata = dict()
    if unpack == "directory":
        dir_to_include = file_item.name if include_dir else None
        dataset = [_build_file_location(unpacked_file, dir_to_include)
            for unpacked_file in unpack_dir(file_item)]
        metadata["dataset"] = dataset
    elif unpack:
        dataset = [_build_file_location(unpacked_file)
            for unpacked_file in unpack_archive(file_item, unpack)]
        metadata["dataset"] = dataset
    else:
        file_location = _build_file_location(file_item)
        metadata.update(file_location)
    return metadata


def unpack_archive(path, unpack):
    """Unpack the path and yield the extracted filenames."""
    fs = fsspec.get_filesystem_class(unpack)(fsspec.open(path.path, **path.storage_options))
    files = fs.find("/")
    for fi in files:
        yield UPath(fi,
                    protocol=unpack,
                    target_protocol=path.protocol,
                    target_options=path.storage_options,
                    fo=as_uri(path))


def unpack_dir(path):
    """Unpack the directory and generate the files it contains (recursively)."""
    files = path.fs.find(path.path)
    for fi in files:
        yield UPath(fi,
                    protocol=path.protocol,
                    **path.storage_options)


def _build_file_location(file_item, include_dir=None):
    file_location = dict()
    try:
        with dummy_connect(file_item):
            file_location["filesystem"] = json.loads(file_item.fs.to_json(include_password=False))

        file_location["uri"] = as_uri(file_item)
        file_location["path"] = file_item.path
    except AttributeError:  # fileitem is not a UPath if it cannot access .fs
        file_location["uri"] = str(file_item)

    if include_dir:
        uid = include_dir + file_item.path.rsplit(include_dir, 1)[-1]
    else:
        uid = file_item.name
    file_location["uid"] = uid
    return file_location


def as_uri(file_item):
    """Represent file item’s path as an unquoted uri."""
    with suppress(AttributeError):
        protocol = file_item.protocol
        if protocol.startswith("http"):
            return file_item.as_uri()
    return unquote(file_item.as_uri())


@contextmanager
def dummy_connect(file_item):
    """Make the _connect method of the fsspec class a no-op.

    This is for the case where only serialization of the filesystem is needed.
    """
    def _fake_connect(*_args, **_kwargs): ...

    klass = fsspec.get_filesystem_class(file_item.protocol)
    try:
        original_connect = klass._connect
    except AttributeError:
        yield
        return

    klass._connect = _fake_connect
    try:
        yield
    finally:
        klass._connect = original_connect


def apply_aliases(aliases, metadata):
    """Apply aliases to the metadata.

    Args:
        aliases: a dict containing dicts for each key to be aliases. For example
            `{"platform_name": {"npp": "Suomi-NPP}}` will replace the `platform_name` "npp" with "Suomi-NPP".
        metadata: the metadata to fix
    """
    for key, val in metadata.items():
        if key in aliases:
            metadata[key] = aliases[key].get(val, val)


def fix_times(info):
    """Fix times so that date and time components are combined, and start time is before end time."""
    if "start_date" in info:
        info["start_time"] = datetime.datetime.combine(info["start_date"].date(),
                                                       info["start_time"].time())
        if "end_date" not in info:
            info["end_date"] = info["start_date"]
        del info["start_date"]
    if "end_date" in info:
        info["end_time"] = datetime.datetime.combine(info["end_date"].date(),
                                                     info["end_time"].time())
        del info["end_date"]
    if "end_time" in info:
        while info["start_time"] > info["end_time"]:
            info["end_time"] += datetime.timedelta(days=1)

def parse_metadata(file_pattern, path):
    """Parse metadata from the filename."""
    if file_pattern is not None:
        file_metadata = parse(file_pattern, path)
        fix_times(file_metadata)
    else:
        file_metadata = {}
    return file_metadata



address = "https://panda.copernicus.eu/"

# def test_one():
#     session = requests.session()
#     session.headers = {
#             "User-Agent": "pytroll/0.0.1",
#         }
#     session.auth = ("martinraspaud","FoHeefe2AeB.")
#     session.get("https://panda.copernicus.eu/Mc3OpenSearch/webapi/Services/getDescriptionFile")


import feedparser
import requests

feed = feedparser.parse("http://panda.copernicus.eu/Mc3OpenSearch/webapi/Services/getProducts/?instrument=SAR&start=2024-07-15T00:00:00.000Z&end=2024-07-30T00:00:00.000Z")
print(f"found {len(feed.entries)} entries for SAR data since july 15th.")
for entry in feed.entries:
    for link in entry["links"]:
        if link["type"] == "application/binary":
            download_link = link["href"]
            filename = entry["id"].rsplit(":", 1)[-1]
            print(f"downloading {filename}")
print(entry)
            #print(f"from {download_link}")

            # with requests.get(download_link, stream=True) as r:
            #     r.raise_for_status()
            #     with open(filename, "wb") as f:
            #         for chunk in r.iter_content(chunk_size=8192):
            #             f.write(chunk)



"""Tests for the gathered publisher functions."""

import json
import logging
from unittest import mock

import pytest
import yaml
from posttroll.message import Message
from posttroll.testing import patched_publisher
from upath import UPath

from pytroll_watchers.local_watcher import file_generator as local_generator
from pytroll_watchers.local_watcher import file_publisher as local_publisher
from pytroll_watchers.main_interface import (
    cli,
    get_generator_for_backend,
    get_publisher_for_backend,
    publish_from_config,
)
from pytroll_watchers.minio_notification_watcher import file_generator as minio_generator
from pytroll_watchers.minio_notification_watcher import file_publisher as minio_publisher
from pytroll_watchers.testing import patched_bucket_listener, patched_local_events  # noqa


def test_getting_right_publisher():
    """Test getting the right publisher for a given backend."""
    file_publisher = get_publisher_for_backend("minio")
    assert file_publisher == minio_publisher
    file_publisher = get_publisher_for_backend("local")
    assert file_publisher == local_publisher
    with pytest.raises(ValueError, match="Unknown backend"):
        _ = get_publisher_for_backend("some_other_backend")

def test_getting_right_generator():
    """Test getting the right generator for a given backend."""
    generator = get_generator_for_backend("minio")
    assert generator == minio_generator
    generator = get_generator_for_backend("local")
    assert generator == local_generator
    with pytest.raises(ValueError, match="Unknown backend"):
        _ = get_generator_for_backend("some_other_backend")


def test_pass_config_to_file_publisher_for_local_backend(tmp_path, patched_local_events):  # noqa
    """Test passing a config to create a file publisher from a local fs."""
    local_settings = dict(directory=tmp_path)
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    config = dict(backend="local",
                  fs_config=local_settings,
                  publisher_config=publisher_settings,
                  message_config=message_settings)
    with patched_publisher() as msgs:
        filename = tmp_path / "bla"
        with patched_local_events([filename]):
            publish_from_config(config)
            assert len(msgs) == 1
            assert str(filename) in msgs[0]


def test_pass_config_to_file_publisher_for_local_backend_with_protocol(tmp_path, patched_local_events, monkeypatch):  # noqa
    """Test passing a config to create a file publisher from a local fs with protocol."""
    new_fs = mock.Mock()
    host = "myhost.pytroll.org"
    fs = dict(cls="fsspec.implementations.sftp.SFTPFileSystem",
              protocol="sftp",
              args=[],
              host=host)
    new_fs.to_json.return_value = json.dumps(fs)
    monkeypatch.setattr(UPath, "fs", new_fs)
    local_settings = dict(directory=tmp_path, protocol="ssh", storage_options=dict(host=host))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    config = dict(backend="local",
                  fs_config=local_settings,
                  publisher_config=publisher_settings,
                  message_config=message_settings)
    with patched_publisher() as msgs:
        filename = tmp_path / "bla"
        with patched_local_events([filename]):
            publish_from_config(config)
            assert len(msgs) == 1
            msg = Message.decode(msgs[0])
            assert msg.data["path"] == str(filename)



def test_pass_config_to_object_publisher_for_minio_backend(patched_bucket_listener):  # noqa
    """Test passing a config to create an objec publisher from minio bucket."""
    s3_settings = dict(endpoint_url="someendpoint",
                       bucket_name="viirs-data",
                       storage_options=dict())
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    config = dict(backend="minio",
                  fs_config=s3_settings,
                  publisher_config=publisher_settings,
                  message_config=message_settings)

    records = [{"Records": [{
               "eventName": "s3:ObjectCreated:Put",
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "object": {"contentType": "application/x-hdf5",
                                 "key": "sdr/bla.h5",
                                 "size": 22183568,
                                 "userMetadata": {"content-type": "application/x-hdf5"}}}}]}]


    with patched_publisher() as msgs:
        with patched_bucket_listener(records):
            publish_from_config(config)
            assert len(msgs) == 1
            assert str("sdr/bla.h5") in msgs[0]


def test_pass_config_to_file_publisher_for_spurious_backend():
    """Test that spurious backend fails."""
    config = {}
    config["backend"] = "some_other_backend"
    with pytest.raises(ValueError, match="Unknown backend"):
        publish_from_config(config)


def test_cli(tmp_path, patched_local_events):  # noqa
    """Test the command-line interface."""
    local_settings = dict(directory=str(tmp_path))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    config = dict(backend="local",
                  fs_config=local_settings,
                  publisher_config=publisher_settings,
                  message_config=message_settings)

    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as fd:
        fd.write(yaml.dump(config))

    with patched_publisher() as msgs:
        filename = tmp_path / "bla"
        with patched_local_events([filename]):
            cli([str(config_file)])
            assert len(msgs) == 1
            assert str(filename) in msgs[0]


def test_cli_with_logging(tmp_path, patched_local_events):  # noqa
    """Test the command-line interface with logging."""
    local_settings = dict(directory=str(tmp_path))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    config = dict(backend="local",
                  fs_config=local_settings,
                  publisher_config=publisher_settings,
                  message_config=message_settings)

    config_file = tmp_path / "config.yaml"
    with open(config_file, "w") as fd:
        fd.write(yaml.dump(config))

    log_config_file = tmp_path / "log_config.yaml"
    handler_name = "console123"
    log_config = {
        "version": 1,
        "handlers": {
            handler_name: {
                "class": "logging.StreamHandler",
                "level": "INFO",
            },
        },
        "loggers": {
            "": {
                "level": "INFO",
                "handlers": [handler_name],
            },
        },
    }
    with open(log_config_file, "w") as fd:
        fd.write(yaml.dump(log_config))

    with patched_publisher():
        filename = tmp_path / "bla"
        with patched_local_events([filename]):
            cli([str(config_file), "-l", str(log_config_file)])
            root = logging.getLogger()
            assert len(root.handlers) == 1
            assert root.handlers[0].name == handler_name



"""Functions and classes for performing message selection.

Selection in this context means making sure only one message refering to some file will be published further.

This is useful when multiple source for the same data are sending messages (eg two reception servers for eumetcast) but
only one of each file is needed for further processing.

To check if two messages refer to the same data, the *uid* metadata of the messages is used.

At the moment, this module makes use of redis as a refined dictionary for keeping track of the received files. Hence a
redis server instance will be started along with some of the functions here, and with the cli.

A command-line script is also made available by this module. It is called ``pytroll-selector``::

    usage: pytroll-selector [-h] [-l LOG_CONFIG] config

    Selects unique messages (based on uid) from multiple sources.

    positional arguments:
      config                The yaml config file.

    options:
      -h, --help            show this help message and exit
      -l LOG_CONFIG, --log-config LOG_CONFIG
                            The yaml config file for logging.

    Thanks for using pytroll-selector!

An example config file to use with this script is the following::

    selector_config:
      ttl: 30
    publisher_config:
      name: hrit_selector
    subscriber_config:
      addresses:
        - tcp://eumetcast_reception_1:9999
        - tcp://eumetcast_reception_2:9999
      nameserver: false
      topics:
        - /1b/hrit-segment/0deg

The different sections are passed straight on to :py:func:`run_selector`, so check it to have more information about
what to pass to it.


"""

import argparse
import logging
import time
from contextlib import closing, contextmanager
from functools import cache
from pathlib import Path
from subprocess import Popen

import redis
import yaml
from posttroll.publisher import create_publisher_from_dict_config
from posttroll.subscriber import create_subscriber_from_dict_config

from pytroll_watchers.main_interface import configure_logging

logger = logging.getLogger(__name__)

@cache
def _connect_to_redis(**kwargs):
    return redis.Redis(**kwargs)


class TTLDict:
    """A simple dictionary-like object that discards items older than a time-to-live.

    Not thread-safe.

    Args:
        ttl: the time to live of the stored items in integer seconds or as a timedelta instance. Cannot be less
            than 1 second.
        redis_params: the keyword arguments to pass to the underlying :py:class:`~redis.Redis` instance.

    """

    def __init__(self, ttl=300, **redis_params):
        """Set up the instance.

        Args:
            ttl: the time to live of the stored items in integer seconds or as a timedelta instance. Cannot be less
              than 1 second.
            redis_params: the keyword arguments to pass to the underlying :py:class:`~redis.Redis` instance.
        """
        self._redis = _connect_to_redis(**redis_params)
        self._ttl = ttl

    def __getitem__(self, key):
        """Get the value corresponding to *key*."""
        return self._redis[key]

    def __setitem__(self, key, value):
        """Set the *value* corresponding to *key*."""
        res = self._redis.get(key)
        if not res:
            self._redis.set(key, value, ex=self._ttl)

    def __contains__(self, key):
        """Check if key is already present."""
        try:
            _ = self[key]
            return True
        except KeyError:
            return False


def running_selector(selector_config, subscriber_config):
    """Generate selected messages.

    The aim of this generator is to skip messages that are duplicates to already processed messages.
    Duplicate in this context means messages referring to the same file (even if stored in different locations).

    Args:
        selector_config: a dictionary of arguments to pass to the underlying redis instance, see
          :py:class:`~redis.Redis`. You can also provide a ttl as an int
          (seconds) or timedelta instance, otherwise it defaults to 300 seconds (5 minutes).
        subscriber_config: a dictionary of arguments to pass to
          :py:func:`~posttroll.subscriber.create_subscriber_from_dict_config`.

    Yields:
        JSON representations of posttroll messages.
    """
    ttl_dict = TTLDict(**selector_config)

    for msg in _data_messages(subscriber_config):
        key = unique_key(msg)
        msg_string = str(msg)

        if key not in ttl_dict:
            ttl_dict[key] = msg_string
            logger.info(f"New content {msg_string}")
            yield msg_string
        else:
            logger.debug(f"Discarded {msg_string}")


def _data_messages(subscriber_config):
    """Generate messages referring to new data from subscriber settings."""
    subscriber = create_subscriber_from_dict_config(subscriber_config)

    with closing(subscriber):
        for msg in subscriber.recv():
            if msg.type != "file":
                continue
            yield msg


def unique_key(msg):
    """Identify the content of the message with a unique key."""
    return msg.data["uid"]


def _run_selector_with_managed_dict_server(selector_config, subscriber_config, publisher_config):
    """Run the selector with a managed ttldict server."""
    publisher = create_publisher_from_dict_config(publisher_config)
    publisher.start()
    with closing(publisher):
        for msg in running_selector(selector_config, subscriber_config):
            publisher.send(msg)


def run_selector(selector_config, subscriber_config, publisher_config):
    """Run the selector.

    The aim of the selector is to skip messages that are duplicates to already published messages.
    Duplicate in this context means messages referring to the same file (even if stored in different locations).

    Messages that refer to new files will be published.

    Args:
        selector_config: a dictionary of arguments to pass to the underlying redis instance, see
          :py:class:`~redis.Redis`. You can also provide a *ttl* for the
          selector as an int (seconds) or timedelta instance, so that incoming messages are forgotten after that time.
          If not provided, the ttl defaults to 300 seconds (5 minutes).
          Also, you can provide a *directory* for the underlying datastructure to store the data in.
        subscriber_config: a dictionary of arguments to pass to
          :py:func:`~posttroll.subscriber.create_subscriber_from_dict_config`. The subscribtion is used as a source for
          messages to process.
        publisher_config: a dictionary of arguments to pass to
          :py:func:`~posttroll.publisher.create_publisher_from_dict_config`. This publisher will send the selected
          messages.

    """
    with _started_redis_server(port=selector_config.get("port"), directory=selector_config.pop("directory", None)):
        _run_selector_with_managed_dict_server(selector_config, subscriber_config, publisher_config)


@contextmanager
def _started_redis_server(port=None, directory=None):
    command = ["redis-server"]
    if port:
        port = str(int(port))  # using int first here prevents arbitrary strings to be passed to Popen
        command += ["--port", port]
    if directory:
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)
        command += ["--dir", directory]
    proc = Popen(command)  # noqa:S603  port is validated and directory is a Path
    time.sleep(.25)
    try:
        yield
    finally:
        proc.terminate()
        proc.wait(3)


def cli(args=None):
    """Command line interface."""
    parser = argparse.ArgumentParser(
                    prog="pytroll-selector",
                    description="Selects unique messages (based on uid) from multiple sources.",
                    epilog="Thanks for using pytroll-selector!")

    parser.add_argument("config", type=str, help="The yaml config file.")
    parser.add_argument("-l", "--log-config", type=str, help="The yaml config file for logging.", default=None)

    parsed = parser.parse_args(args)


    log_config_filename = parsed.log_config
    configure_logging(log_config_filename)

    config_file = parsed.config

    with open(config_file) as fd:
        config_dict = yaml.safe_load(fd.read())

    selector_config = config_dict.get("selector_config", {})
    subscriber_config = config_dict.get("subscriber_config", {})
    publisher_config = config_dict.get("publisher_config", {})

    return run_selector(selector_config, subscriber_config, publisher_config)



"""Module to provide file generator and publisher for the Copernicus dataspace contents.

It polls the catalogue using OData for new data (https://documentation.dataspace.copernicus.eu/APIs/OData.html) and
generates locations for the data on the S3 services (https://documentation.dataspace.copernicus.eu/APIs/S3.html).

Note:
    The OData and S3 services require two different set of credentials.


Example of configuration file to retrieve SAR data from dataspace:

.. code-block:: yaml


    backend: dataspace
    fs_config:
      filter_string: "contains(Name,'IW_GRDH')"
      dataspace_auth:
        netrc_host: catalogue.dataspace.copernicus.eu
      storage_options:
        profile: copernicus
      polling_interval:
        minutes: 10
      start_from:
        hours: 1
    publisher_config:
      name: s1_watcher
      nameservers: false
      port: 3000
    message_config:
      subject: /segment/s1/l1b/
      atype: file
      aliases:
        sensor:
          SAR: SAR-C
    data_config:
      unpack:
        format: directory
        include_dir_in_uid: true
"""

import datetime
import logging
import netrc
import time
from contextlib import suppress
from functools import lru_cache

from oauthlib.oauth2 import LegacyApplicationClient
from requests_oauthlib import OAuth2Session
from upath import UPath

from pytroll_watchers.common import fromisoformat, run_every
from pytroll_watchers.publisher import file_publisher_from_generator

client_id = "cdse-public"
token_url = "https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token"  # noqa
logger = logging.getLogger(__name__)


def file_publisher(config):
    """Publish files coming from local filesystem events.

    Args:
        config: the configuration dictionary, containing in particular an fs_config section, which is the configuration
        for the filesystem watching, will be passed as argument to `file_generator`. The other sections are passed
        further to ``file_publisher_from_generator``.
    """
    fs_config = config["fs_config"]
    logger.info(f"Starting watch on dataspace for '{fs_config['filter_string']}'")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, config)


def file_generator(filter_string,
                   polling_interval,
                   dataspace_auth,
                   start_from=None,
                   storage_options=None):
    """Generate new objects by polling copernicus dataspace.

    Args:
        filter_string: the filter to use for narrowing the data to poll. For example, to poll level 1 olci data, it can
            be `contains(Name,'OL_1_EFR')`. For more information of the filter parameters, check:
            https://documentation.dataspace.copernicus.eu/APIs/OData.html
        polling_interval: the interval (timedelta object or kwargs to timedelta) at which the dataspace will be polled.
        dataspace_auth: the authentication information, as a dictionary. It can be a dictionary with `username` and
            `password` keys, or with `netrc_host` and optionaly `netrc_file` if the credentials are to be fetched with
            netrc.
        start_from: how far back in time to fetch the data the first time. This is helpful for the first iteration of
            the generator, so that data from the past can be fetched, to fill a possible gap. Default to 0, meaning
            nothing older than when the generator starts will be fetched. Same format accepted as polling_interval.
        storage_options: The options to pass the S3Path instance, usually include ways to get credentials to the
            copernicus object store, like `profile` from the .aws configuration files.

    Yields:
        Tuples of UPath (s3) and metadata.

    """
    with suppress(TypeError):
        polling_interval = datetime.timedelta(**polling_interval)
    with suppress(TypeError):
        start_from = datetime.timedelta(**start_from)
    if start_from is None:
        start_from = datetime.timedelta(0)

    last_pub_date = datetime.datetime.now(datetime.timezone.utc) - start_from

    for next_check in run_every(polling_interval):
        generator = generate_download_links_since(filter_string, dataspace_auth, last_pub_date, storage_options)
        for s3path, metadata in generator:
            last_pub_date = _update_last_publication_date(last_pub_date, metadata)
            yield s3path, metadata
        logger.info("Finished polling.")
        if next_check > datetime.datetime.now(datetime.timezone.utc):
            logger.info(f"Next iteration at {next_check}")


def _update_last_publication_date(last_publication_date, metadata):
    """Update the last publication data based on the metadata."""
    publication_date = fromisoformat(metadata.pop("PublicationDate"))
    if publication_date > last_publication_date:
        last_publication_date = publication_date
    return last_publication_date


def generate_download_links_since(filter_string, dataspace_auth, last_publication_date, storage_options):
    """Generate download links for data that was published since a given `last publication_date`.

    Example:
        To fetch download link since yesterday, using netrc-stored credentials, and an aws s3 profile:

        >>> from pytroll_watchers.dataspace_watcher import generate_download_links_since
        >>> filter_string = "contains(Name,'OL_1_EFR')"
        >>> dataspace_auth = dict(netrc_host="dataspace.copernicus.eu")
        >>> last_publication_date = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(hours=24)
        >>> storage_options = dict(profile="my_copernicus_s3_profile")
        >>> generator = generate_download_links_since(filter_string, dataspace_auth, last_publication_date,
        ...                                           storage_options)

    """
    pub_limit = f"PublicationDate gt {last_publication_date.strftime('%Y-%m-%dT%H:%M:%S.%fZ')}"
    filter_string_with_pub_limit = f"{filter_string} and {pub_limit}"

    yield from generate_download_links(filter_string_with_pub_limit, dataspace_auth, storage_options)


def generate_download_links(filter_string, dataspace_auth, storage_options):
    """Generate download links for a given filter_string."""
    if storage_options is None:
        storage_options = {}
    oauth = _get_auth(**dataspace_auth)
    resp = oauth.get(filter_string)
    metadatas = resp.get("value", [])
    for metadata in metadatas:
        s3path = UPath("s3:/" + metadata["S3Path"], **storage_options)
        mda = dict()
        attributes = _construct_attributes_dict(metadata)
        mda["platform_name"] = attributes["platformShortName"].capitalize() + attributes["platformSerialIdentifier"]
        mda["sensor"] = attributes["instrumentShortName"].lower()
        mda["PublicationDate"] = metadata["PublicationDate"]
        mda["boundary"] = metadata["GeoFootprint"]
        mda["product_type"] = attributes["productType"]
        mda["start_time"] = fromisoformat(attributes["beginningDateTime"])
        mda["end_time"] = fromisoformat(attributes["endingDateTime"])
        mda["orbit_number"] = int(attributes["orbitNumber"])

        for checksum in metadata["Checksum"]:
            if checksum.get("Algorithm") == "MD5":
                mda["checksum"] = dict(algorithm=checksum["Algorithm"], hash=checksum["Value"])
                break
        mda["size"] = int(metadata["ContentLength"])

        yield s3path, mda


def _construct_attributes_dict(entry):
    """Construct a dict from then "results" item in entry."""
    results = entry["Attributes"]
    attributes = {result["Name"]: result["Value"] for result in results}
    return attributes


@lru_cache(maxsize=1)
def _get_auth(**dataspace_auth):
    oauth = CopernicusOAuth2Session(dataspace_auth)
    return oauth


class CopernicusOAuth2Session():
    """An oauth2 session for copernicus dataspace."""


    def __init__(self, dataspace_auth):
        """Set up the session."""
        dataspace_credentials = _get_credentials(dataspace_auth)
        self._oauth = OAuth2Session(client=LegacyApplicationClient(client_id=client_id))
        def compliance_hook(response):
            response.raise_for_status()
            return response

        self._oauth.register_compliance_hook("access_token_response", compliance_hook)
        self._token_user, self._token_pass = dataspace_credentials

    def get(self, filter_string):
        """Run a get request."""
        self.fetch_token()
        url = f"https://catalogue.dataspace.copernicus.eu/odata/v1/Products?$filter={filter_string}&$expand=Attributes"
        return self._oauth.get(url).json()

    def fetch_token(self):
        """Fetch the token."""
        if not self._oauth.token or self._oauth.token["expires_at"] <= time.time():
            self._oauth.fetch_token(token_url=token_url,
                                    username=self._token_user,
                                    password=self._token_pass)


def _get_credentials(ds_auth):
    """Get credentials from the ds_auth dictionary."""
    try:
        creds = ds_auth["username"], ds_auth["password"]
    except KeyError:
        username, _, password = netrc.netrc(ds_auth.get("netrc_file")).authenticators(ds_auth["netrc_host"])
        creds = (username, password)
    return creds



"""Module to provide file generator and publisher for the EUMETSAT datastore contents.

It polls the catalogue using Opensearch for new data and generates locations for the data on https.

Note:
    The links produced can only be downloaded with a valid token. A token comes with the links, but
    has only a limited validity time (maybe 5 minutes).

An example for getting links to MSG data::

    from pytroll_watchers.datastore_watcher import generate_download_links_since

    ds_auth = dict(netrc_host="api.eumetsat.int")

    collection = "EO:EUM:DAT:MSG:HRSEVIRI"

    search_params = dict(collection=collection)
    now = datetime.datetime.now(datetime.timezone.utc)
    yesterday = now - datetime.timedelta(hours=6)

    links = list(generate_download_links_since(search_params, ds_auth, yesterday))

Another example, here a configuration file to pass to the CLI::

    backend: datastore
    fs_config:
      search_params:
        collection: "EO:EUM:DAT:0905"
      polling_interval:
        minutes: 5
      start_from:
        hours: 6
      ds_auth:
        netrc_host: api.eumetsat.int

    publisher_config:
      name: <my watcher publisher name>
      nameservers: false
      port: <my port number>
    message_config:
      subject: /my/datastore/watcher/topic
      atype: file
      data:
        sensor: aws
        platform_name: AWS1
        variant: GDS
"""

import datetime
import logging
import netrc
import time
from contextlib import suppress

from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from upath import UPath

from pytroll_watchers.common import fromisoformat, run_every
from pytroll_watchers.publisher import file_publisher_from_generator
from pytroll_watchers.version import version

logger = logging.getLogger(__name__)

token_url = "https://api.eumetsat.int/token"  # noqa
data_url = "https://api.eumetsat.int/data"



def file_publisher(config):
    """Publish files coming from local filesystem events.

    Args:
        config: the configuration dictionary, containing in particular an fs_config section, which is the configuration
        for the filesystem watching, will be passed as argument to `file_generator`. The other sections are passed
        further to ``file_publisher_from_generator``.
    """
    fs_config = config["fs_config"]
    logger.info(f"Starting watch on datastore for '{fs_config['search_params']}'")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, config)


def file_generator(search_params, polling_interval, ds_auth, start_from=None):
    """Search params must contain at least collection.

    Args:
        search_params: the dictionary of search parameters to request. Based on the opensearch API:
          https://user.eumetsat.int/api-definitions/data-store-opensearch-api
        polling_interval: how often to poll for new data. Can be provided as a timedelta or a dictionary of arguments
          for timedelta.
        ds_auth: either a dictionary with `netrc_host` (and optionally `netrc_file`), or a dictionary with `username`
          and `password`.
        start_from: a timedelta or dictionary of arguments to timedelta to specify how far in time to start fetching
          data. `None` by default, which means the data will be no older that now.
    """
    with suppress(TypeError):
        polling_interval = datetime.timedelta(**polling_interval)
    with suppress(TypeError):
        start_from = datetime.timedelta(**start_from)
    if start_from is None:
        start_from = datetime.timedelta(0)

    last_pub_date = datetime.datetime.now(datetime.timezone.utc) - start_from
    for next_check in run_every(polling_interval):
        new_pub_date =  datetime.datetime.now(datetime.timezone.utc)
        yield from generate_download_links_since(search_params, ds_auth, last_pub_date)
        logger.info("Finished polling.")
        if next_check > datetime.datetime.now(datetime.timezone.utc):
            logger.info(f"Next iteration at {next_check}")
            last_pub_date = new_pub_date

def generate_download_links_since(search_params, ds_auth, start_from):
    """Generate download links for data that was published since `start_from`."""
    str_pub_start = start_from.isoformat(timespec="milliseconds")
    search_params = search_params.copy()
    search_params["publication"] = f"[{str_pub_start}"
    yield from generate_download_links(search_params, ds_auth)


def generate_download_links(search_params, ds_auth):
    """Generate download links provide search parameter and authentication."""
    session = DatastoreOAuth2Session(ds_auth)
    collection = search_params.pop("collection")
    request_params = {
            "format": "json",
            "pi": str(collection),
            "si": 0,
            "c": 100,  # items per page
        }

    if search_params:
        request_params.update(search_params)

    jres = session.get(request_params)
    headers={"Authorization": f"Bearer {session.token['access_token']}"}
    client_args = dict(headers=headers)
    for feature in jres["features"]:
        links = feature["properties"]["links"]["data"]
        if len(links) != 1:
            raise ValueError("Don't know how to generate multiple files at the time.")
        path = UPath(links[0]["href"], encoded=True, client_kwargs=client_args)
        # In the future, it might be interesting to generate items from the sip-entries, as
        # they contain individual files for zip archives.
        mda = dict()
        mda["boundary"] = feature["geometry"]
        acq_info = feature["properties"]["acquisitionInformation"][0]
        mda["platform_name"] = acq_info["platform"]["platformShortName"]
        mda["sensor"] = acq_info["instrument"]["instrumentShortName"].lower()
        with suppress(KeyError):
            mda["orbit_number"] = acq_info["acquisitionParameters"]["orbitNumber"]
        start_string, end_string = feature["properties"]["date"].split("/")
        mda["start_time"] = fromisoformat(start_string)
        mda["end_time"] = fromisoformat(end_string)
        mda["product_type"] = str(collection)
        mda["checksum"] = dict(algorithm="md5", hash=feature["properties"]["extraInformation"]["md5"])
        yield path, mda


class DatastoreOAuth2Session():
    """An oauth2 session for eumetsat datastore."""

    def __init__(self, datastore_auth):
        """Set up the session."""
        client_id, client_secret = _get_credentials(datastore_auth)
        self._oauth = OAuth2Session(client=BackendApplicationClient(client_id=client_id))
        def compliance_hook(response):
            response.raise_for_status()
            return response

        self._oauth.register_compliance_hook("access_token_response", compliance_hook)
        self._token_secret = client_secret

    def get(self, params):
        """Run a get request."""
        self.fetch_token()
        search_url = f"{data_url}/search-products/1.0.0/os"
        headers = {"referer": "https://github.com/pytroll/pytroll-watchers",
                   "User-Agent": f"pytroll-watchers / {version}"}

        return self._oauth.get(search_url, params=params, headers=headers).json()

    @property
    def token(self):
        """Return the current token."""
        return self.fetch_token()

    def fetch_token(self):
        """Fetch the token."""
        if not self._oauth.token or self._oauth.token["expires_at"] <= time.time():
            self._oauth.fetch_token(token_url=token_url,
                                    client_secret=self._token_secret,
                                    include_client_id=True)
        return self._oauth.token


def _get_credentials(ds_auth):
    """Get credentials from the ds_auth dictionary."""
    try:
        creds = ds_auth["username"], ds_auth["password"]
    except KeyError:
        username, _, password = netrc.netrc(ds_auth.get("netrc_file")).authenticators(ds_auth["netrc_host"])
        creds = (username, password)
    return creds



"""Watcher for DHuS instances.

For more information about DHuS, check out
https://sentineldatahub.github.io/DataHubSystem/about.html


An example configuration file to retrieve Sentinel 1 data from a DHuS instance:

.. code-block:: yaml

  backend: dhus
  fs_config:
    server: https://myhub.someplace.org/
    filter_params:
        - substringof('IW_GRDH',Name)
    polling_interval:
        seconds: 10
    start_from:
        hours: 6
  publisher_config:
    name: s1_watcher
  message_config:
    subject: /segment/s1/l1b/
    atype: dataset
    aliases:
        sensor:
          SAR: SAR-C
  data_config:
    unpack:
      format: zip
"""

import datetime as dt
import logging
from contextlib import suppress
from urllib.parse import urljoin

import requests
from geojson import Polygon
from upath import UPath

from pytroll_watchers.common import fromisoformat, run_every
from pytroll_watchers.publisher import file_publisher_from_generator

logger = logging.getLogger(__name__)


def file_publisher(config):
    """Publish files coming from local filesystem events.

    Args:
        config: the configuration dictionary, containing in particular an fs_config section, which is the configuration
        for the filesystem watching, will be passed as argument to `file_generator`. The other sections are passed
        further to ``file_publisher_from_generator``.
    """
    fs_config = config["fs_config"]
    logger.info(f"Starting watch on dhus for '{fs_config['filter_params']}'")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, config)


def file_generator(server, filter_params, polling_interval, start_from=None):
    """Generate new objects by polling a DHuS instance.

    Args:
        server: the DHuS server to use.
        filter_params: the list of filter parameters to use for narrowing the data to poll. For example, to poll IW sar
            data, it can be `substringof('IW_GRDH',Name)`. For more information of the filter parameters, check:
            https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/ODataAPI#filter
        polling_interval: the interval (timedelta object or kwargs to timedelta) at which the DHUS will be polled.
        start_from: how far back in time to fetch the data the first time. This is helpful for the first iteration of
            the generator, so that data from the past can be fetched, to fill a possible gap. Default to 0, meaning
            nothing older than when the generator starts will be fetched. Same format accepted as polling_interval.

    Yields:
        Tuples of UPath (http) and metadata.


    Note:
        As this watcher uses requests, the authentication information should be stored in a .netrc file.
    """
    with suppress(TypeError):
        polling_interval = dt.timedelta(**polling_interval)
    with suppress(TypeError):
        start_from = dt.timedelta(**start_from)
    if start_from is None:
        start_from = dt.timedelta(0)

    last_pub_date = dt.datetime.now(dt.timezone.utc) - start_from
    for next_check in run_every(polling_interval):
        generator = generate_download_links_since(server, filter_params, last_pub_date)
        for path, metadata in generator:
            last_pub_date = _update_last_publication_date(last_pub_date, metadata)
            yield path, metadata
        logger.info("Finished polling.")
        if next_check > dt.datetime.now(dt.timezone.utc):
            logger.info(f"Next iteration at {next_check}")


def _update_last_publication_date(last_publication_date, metadata):
    """Update the last publication data based on the metadata."""
    publication_date = metadata.pop("ingestion_date")
    if publication_date > last_publication_date:
        last_publication_date = publication_date
    return last_publication_date


def generate_download_links_since(server, filter_params, last_publication_date):
    """Generate download links for the data published since `last_publication_date`."""
    last_publication_date = last_publication_date.astimezone(dt.timezone.utc)
    # remove timezone info as dhus considers all times utc
    pub_string = last_publication_date.isoformat(timespec="milliseconds")[:-6]
    filter_params = filter_params + [f"IngestionDate gt datetime'{pub_string}'"]
    yield from generate_download_links(server, filter_params)


def generate_download_links(server, filter_params):
    """Generate download links.

    The filter params we can use are defined here: https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/ODataAPI#filter
    """
    filter_string = " and ".join(filter_params)

    url = urljoin(server, f"/odata/v1/Products?$format=json&$filter={filter_string}&$expand=Attributes")

    response = requests.get(url, timeout=60)
    response.raise_for_status()

    entries = response.json()

    for entry in entries["d"]["results"]:
        mda = dict()
        path = UPath(entry["__metadata"]["media_src"], client_kwargs=dict(trust_env=True))
        mda["boundary"] = _extract_boundary_as_geojson(entry)
        attributes = _construct_attributes_dict(entry)
        mda["platform_name"] = attributes["Satellite name"].capitalize() + attributes["Satellite number"]
        mda["sensor"] = attributes["Instrument"]
        mda["ingestion_date"] = fromisoformat(attributes["Ingestion Date"])
        mda["product_type"] = attributes["Product type"]
        mda["start_time"] = fromisoformat(attributes["Sensing start"])
        mda["end_time"] = fromisoformat(attributes["Sensing stop"])
        mda["orbit_number"] = int(attributes["Orbit number (start)"])

        mda["checksum"] = dict(algorithm=entry["Checksum"]["Algorithm"], hash=entry["Checksum"]["Value"])
        mda["size"] = int(entry["ContentLength"])
        mda["uid"] = attributes["Filename"]
        yield path, mda

def _construct_attributes_dict(entry):
    """Construct a dict from then "results" item in entry."""
    results = entry["Attributes"]["results"]
    results_dict = {result["Name"]: result["Value"] for result in results}
    return results_dict


def _extract_boundary_as_geojson(entry):
    """Extract the boundary from the entry metadata."""
    gml, nsmap = read_gml(entry["ContentGeometry"])
    boundary_text = gml.find("gml:outerBoundaryIs/gml:LinearRing/gml:coordinates", namespaces=nsmap).text
    boundary_list = (coords.split(",") for coords in boundary_text.strip().split(" "))
    boundary = Polygon([[(float(lon), float(lat)) for (lat, lon) in boundary_list]])
    return boundary

def read_gml(gml_string):
    """Read the gml string."""
    from xml.etree import ElementTree

    from defusedxml.ElementTree import DefusedXMLParser
    parser = ElementTree.XMLPullParser(["start-ns", "end"], _parser=DefusedXMLParser())
    parser.feed(gml_string)

    nsmap = dict()

    for event, elem in parser.read_events():
        if event == "start-ns":
            nsmap[elem[0]] = elem[1]
    return elem, nsmap



"""Test intermediate steps like fetching and unpacking."""

import os

from posttroll.message import Message
from posttroll.testing import patched_publisher

from pytroll_watchers import local_watcher
from pytroll_watchers.testing import patched_local_events  # noqa


def test_publish_paths_with_fetching(tmp_path, patched_local_events):  # noqa
    """Test publishing paths."""
    basename = "foo+bar,baz_.txt"
    filename = os.fspath(tmp_path / basename)
    with open(filename, "w"):
        pass

    destination = tmp_path / "downloaded"
    destination.mkdir()

    local_settings = dict(directory=tmp_path)
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    data_config = dict(fetch=dict(destination=destination))

    with patched_local_events([filename]):
        with patched_publisher() as messages:
            local_watcher.file_publisher(dict(fs_config=local_settings,
                                              publisher_config=publisher_settings,
                                              message_config=message_settings,
                                              data_config=data_config))

    assert "uri" not in message_settings["data"]
    assert len(messages) == 1
    message = Message(rawstr=messages[0])
    assert message.data["uri"] == f"{str(destination)}/{basename}"
    assert message.data["sensor"] == "viirs"
    assert "filesystem" not in message.data



from urllib.parse import urlencode, urlsplit, urlunsplit

from responses import _recorder
import responses
import requests


opensearch_url = "http://panda.copernicus.eu/Mc3OpenSearch/webapi/Services/getProducts/"

@responses.activate
def test_list():
    filter_params = dict(instrument="SAR", start="2024-11-14T00:00:00.000Z")
    query_string = urlencode(filter_params)
    url_bits = urlsplit(opensearch_url)
    url = urlunsplit((url_bits.scheme, url_bits.netloc, url_bits.path, query_string, ""))
    responses._add_from_file(file_path="tests/panda_search_response.yaml")
    feed_text = requests.get(url, timeout=60).text
    import feedparser

    feed = feedparser.parse(feed_text)
    assert len(feed.entries) == 6



"""Test watching the EUMETSAT Datastore."""

import datetime
from contextlib import contextmanager

import pytest
import responses
import yaml
from freezegun import freeze_time
from posttroll.message import Message
from posttroll.testing import patched_publisher

from pytroll_watchers.datastore_watcher import (
    file_generator,
    file_publisher,
    generate_download_links,
    generate_download_links_since,
)


@pytest.fixture
def search_params():
    """Generate the search parameters for the tests."""
    polygon = ("POLYGON((9.08 60.00,16.25 59.77,17.50 63.06,21.83 66.02,26.41 65.75,22.22 60.78,28.90 60.82,30.54 60.01"
               ",20.20 53.57,9.06 53.50,9.08 60.00))")
    collection = "EO:EUM:DAT:0407"
    return dict(collection=collection, geo=polygon)


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_datastore_get_download_links_since(search_params):
    """Test downloading links from the datastore."""
    ds_auth = dict(username="user", password="pass")  # noqa

    now = datetime.datetime.now(datetime.timezone.utc)
    yesterday = now - datetime.timedelta(hours=24)

    str_pub_start = yesterday.isoformat(timespec="milliseconds")


    response_file = "tests/datastore_responses.yaml"


    with loaded_responses(str_pub_start, response_file):

        search_params = dict(collection=search_params["collection"], geo=search_params["geo"])

        features = generate_download_links_since(search_params, ds_auth, yesterday)
        features = list(features)

    expected_token = "eceba4e1-95e6-3526-8c42-c3c9dc14ff5c"  # noqa

    assert len(features) == 5
    path, mda = features[0]
    assert str(path) == "https://api.eumetsat.int/data/download/1.0.0/collections/EO%3AEUM%3ADAT%3A0407/products/S3B_OL_2_WFR____20240416T104217_20240416T104517_20240417T182315_0180_092_051_1980_MAR_O_NT_003.SEN3"
    assert expected_token in path.storage_options["client_kwargs"]["headers"]["Authorization"]
    assert mda["boundary"]["coordinates"][0][0] == [-14.3786, 52.4516]
    assert mda["platform_name"] == "Sentinel-3B"
    assert mda["sensor"] == "olci"
    assert mda["start_time"] == datetime.datetime(2024, 4, 16, 10, 42, 16, 954262, tzinfo=datetime.timezone.utc)
    assert mda["end_time"] == datetime.datetime(2024, 4, 16, 10, 45, 16, 954262, tzinfo=datetime.timezone.utc)
    assert mda["orbit_number"] == 31123
    assert mda["product_type"] == "EO:EUM:DAT:0407"
    assert mda["checksum"] == dict(algorithm="md5", hash="9057eb08f2a4e9f4c5a8d2eeaacedaef")


@contextmanager
def loaded_responses(since, response_file):
    """Load prerecorded responses for querying the datastore."""
    with responses.RequestsMock() as rsps:
        with open(response_file) as fd:
            contents = yaml.safe_load(fd.read())
            for response in contents["responses"]:
                resp = response["response"]
                resp["url"] = resp["url"].replace("{yesterday}", since.replace("+", "%2B"))
                rsps.add(**response["response"])

        yield


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_datastore_file_generator(tmp_path, search_params):
    """Test the file generator."""
    netrc_host = "myitem"
    netrc_file = tmp_path / "netrc"

    with open(netrc_file, "w") as fd:
        fd.write(f"machine {netrc_host} login user@pytroll.org password mypassword")

    ds_auth = dict(netrc_host=netrc_host, netrc_file=netrc_file)


    now =datetime.datetime.now(datetime.timezone.utc)
    yesterday = now - datetime.timedelta(hours=24)
    start_from = dict(hours=24)


    polling_interval = datetime.timedelta(0)


    str_yesterday = yesterday.isoformat(timespec="milliseconds")
    response_file = "tests/datastore_responses.yaml"

    with loaded_responses(str_yesterday, response_file):
        features = file_generator(search_params, polling_interval, ds_auth, start_from)
        features = list(features)

    expected_token = "eceba4e1-95e6-3526-8c42-c3c9dc14ff5c"  # noqa

    assert len(features) == 5
    path, _ = features[0]
    assert str(path) == "https://api.eumetsat.int/data/download/1.0.0/collections/EO%3AEUM%3ADAT%3A0407/products/S3B_OL_2_WFR____20240416T104217_20240416T104517_20240417T182315_0180_092_051_1980_MAR_O_NT_003.SEN3"
    assert expected_token in path.storage_options["client_kwargs"]["headers"]["Authorization"]


@pytest.fixture
def search_params_geo():
    """Generate the search parameters for the tests."""
    collection = "EO:EUM:DAT:MSG:HRSEVIRI"
    publication_start = datetime.datetime(2023, 11, 11, 8, 0)
    publication_end = datetime.datetime(2023, 11, 11, 8, 15)
    str_pub_start = publication_start.isoformat(timespec="milliseconds")
    str_pub_end = publication_end.isoformat(timespec="milliseconds")
    return dict(collection=collection, publication=f"[{str_pub_start},{str_pub_end}]")


@responses.activate
def test_datastore_generate_download_links_geo(tmp_path, search_params_geo):
    """Test the file generator."""
    netrc_host = "myitem"
    netrc_file = tmp_path / "netrc"

    with open(netrc_file, "w") as fd:
        fd.write(f"machine {netrc_host} login user@pytroll.org password mypassword")

    ds_auth = dict(netrc_host=netrc_host, netrc_file=netrc_file)

    response_file = "tests/datastore_responses_geo.yaml"
    responses._add_from_file(file_path=response_file)

    features = list(generate_download_links(search_params_geo, ds_auth))

    assert len(features) == 1

    path, mda = features[0]
    assert "orbit_number" not in mda
    assert str(path) == "https://api.eumetsat.int/data/download/1.0.0/collections/EO%3AEUM%3ADAT%3AMSG%3AHRSEVIRI/products/MSG3-SEVI-MSG15-0100-NA-20231111081242.335000000Z-NA"
    assert mda["start_time"] == datetime.datetime(2023, 11, 11, 8, 0, 9, 969000, tzinfo=datetime.timezone.utc)
    assert mda["end_time"] == datetime.datetime(2023, 11, 11, 8, 12, 42, 335000, tzinfo=datetime.timezone.utc)



@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_publish_paths(caplog, search_params):
    """Test publishing paths."""
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l2/", atype="file", data=dict(sensor="olci"))


    now =datetime.datetime.now(datetime.timezone.utc)
    yesterday = now - datetime.timedelta(hours=24)
    str_yesterday = yesterday.isoformat(timespec="milliseconds")
    response_file = "tests/datastore_responses.yaml"

    caplog.set_level("INFO")
    with patched_publisher() as messages:
        with loaded_responses(str_yesterday, response_file):
            ds_config = dict(search_params=search_params,
                             polling_interval=dict(minutes=0),
                             start_from=dict(hours=24),
                             ds_auth=dict(username="user5", password="pass5"))  # noqa


            file_publisher(dict(fs_config=ds_config,
                                publisher_config=publisher_settings,
                                message_config=message_settings))
    assert "uri" not in message_settings["data"]
    assert len(messages) == 5
    message = Message(rawstr=messages[0])
    uri = "https://api.eumetsat.int/data/download/1.0.0/collections/EO%3AEUM%3ADAT%3A0407/products/S3B_OL_2_WFR____20240416T104217_20240416T104517_20240417T182315_0180_092_051_1980_MAR_O_NT_003.SEN3"
    assert message.data["uri"] == uri
    assert message.data["sensor"] == "olci"
    assert message.data["filesystem"] == {
        "cls": "fsspec.implementations.http:HTTPFileSystem",
        "protocol": "https",
        "args": [],
        "encoded": True,
        "client_kwargs": {"headers": {"Authorization": "Bearer eceba4e1-95e6-3526-8c42-c3c9dc14ff5c"}},
    }
    assert message.data["path"] == uri
    assert f"Starting watch on datastore for '{search_params}'" in caplog.text



"""Watcher for non-remote file systems.

Either using OS-based envents (like inotify on linux), or polling.

An example configuration file to retrieve data from a directory.

By default, files uris will not include any protocol, which means they will look like `/tmp/myfile`. If it is desired,
the `protocol` setting in the `fs_config` can be set to `"file"` to make the uris look like `file:///tmp/myfile`.

It is also possible to make the local files sent as remote with the `protocol` and `storage_options` settings in the
`fs_config` section. The generated uri can thus start with `ssh://myhost` for example, by setting `protocol: ssh` and
`storage_options: {host: "myhost"}`.


.. code-block:: yaml

  backend: local
  fs_config:
    directory: /data
    file pattern: "H-000-{orig_platform_name:4s}__-{orig_platform_name:4s}_{service:3s}____-{channel_name:_<9s}-\
        {segment:_<9s}-{start_time:%Y%m%d%H%M}-{compression:1s}_"
  publisher_config:
    name: hrit_watcher
  message_config:
    subject: /segment/hrit/l1b/
    atype: file

"""
import logging
import os
from pathlib import Path
from urllib.parse import urlunparse

from upath import UPath

from pytroll_watchers.backends.local import listen_to_local_events
from pytroll_watchers.publisher import SecurityError, file_publisher_from_generator, parse_metadata

logger = logging.getLogger(__name__)


def file_publisher(config):
    """Publish files coming from local filesystem events.

    Args:
        config: the configuration dictionary, containing in particular an fs_config section, which is the configuration
        for the filesystem watching, will be passed as argument to `file_generator`. The other sections are passed
        further to ``file_publisher_from_generator``.
    """
    fs_config = config["fs_config"]
    logger.info(f"Starting watch on '{fs_config['directory']}'")
    if "password" in fs_config.get("storage_options", []):
        raise SecurityError("A password cannot be published safely.")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, config)


def file_generator(directory, observer_type="os", file_pattern=None, protocol=None, storage_options=None):
    """Generate new files appearing in the watched directory.

    Args:
        directory: The locally accessible directory to watch for changes.
        observer_type: What to use for detecting changes.
            It can be either "os" for os-based detections (eg inotify on linux, but can be polling on some os's),
            "polling" for detecting changes through polling, or the actual watchdog class to use as observer.
            Defaults to "os".
        file_pattern: The trollsift pattern to use for matching and extracting metadata from the filename.
            This must not include the directory.
        protocol (optional): In case the file has to be advertised with another protocol than "file".
        storage_options: The storage options for the other protocol. Will be ignored if protocol is None.

    Returns:
        A tuple of Path or UPath and file metadata.

    Examples:
        To iterate over new files in `/tmp/`:

        >>> for filename in file_generator("/tmp/", file_pattern="{start_time:%Y%m%d_%H%M}_{product}.tif")
        ...    print(filename)
        Path("/tmp/20200428_1000_foo.tif")


        To get UPaths with ssh as protocol and a specific host:

        >>> for filename in file_generator("/tmp/", file_pattern="{start_time:%Y%m%d_%H%M}_{product}.tif",
        ...                                protocol="ssh", storage_option=dict(host="myhost.pytroll.org"))
        UPath("ssh:///tmp/20200428_1000_foo.tif")  # .storage_options will show the host.

    """
    file_metadata = {}
    pattern = os.path.join(directory, file_pattern) if file_pattern is not None else None
    with listen_to_local_events(directory, file_pattern, observer_type) as events:
        for path in events:
            try:
                file_metadata = parse_metadata(pattern, path)
            except ValueError:
                continue
            if protocol is not None:
                uri = urlunparse((protocol, None, str(path), None, None, None))
                if storage_options is None:
                    storage_options = dict()
                yield UPath(uri, **storage_options), file_metadata
            else:
                yield Path(path), file_metadata



"""Tests for the publisher functionality."""

import os
from shutil import make_archive

import pytest
from posttroll.message import Message
from posttroll.testing import patched_publisher
from upath import UPath

from pytroll_watchers.publisher import file_publisher_from_generator


def test_unpacking_raises_warning_when_passed_via_message_config(tmp_path):
    """Test unpacking the watched file's components to a dataset message."""
    file1 = tmp_path / "to_zip" / "file1"
    file2 = tmp_path / "to_zip" / "file2"
    zip_file = tmp_path / "archived.zip"

    file1.parent.mkdir()

    open(file1, "a").close()
    open(file2, "a").close()

    make_archive(os.path.splitext(zip_file)[0], "zip", tmp_path / "to_zip")

    path = UPath("file://" + str(zip_file))

    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l2/", atype="dataset", data=dict(sensor="olci"),
                            unpack="zip")

    with patched_publisher():
        with pytest.deprecated_call(match="data_config"):
            file_publisher_from_generator([[path, dict()]],
                                          dict(publisher_config=publisher_settings,
                                               message_config=message_settings))


def test_unpacking(tmp_path):
    """Test unpacking the watched file's components to a dataset message."""
    file1 = tmp_path / "to_zip" / "file1"
    file2 = tmp_path / "to_zip" / "file2"
    zip_file = tmp_path / "archived.zip"

    file1.parent.mkdir()

    open(file1, "a").close()
    open(file2, "a").close()

    make_archive(os.path.splitext(zip_file)[0], "zip", tmp_path / "to_zip")

    path = UPath("file://" + str(zip_file))

    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l2/", atype="dataset", data=dict(sensor="olci"))
    data_config = dict(unpack=dict(format="zip"))

    with patched_publisher() as messages:
        file_publisher_from_generator([[path, dict()]],
                                      dict(publisher_config=publisher_settings,
                                           message_config=message_settings,
                                           data_config=data_config))

    assert "zip://file1" in messages[0]
    assert "zip://file2" in messages[0]


def test_unpacking_directory(tmp_path):
    """Test unpacking the watched directory's components to a dataset message."""
    file1 = tmp_path / "my_dir" / "file1"
    file2 = tmp_path / "my_dir" / "file2"

    dirpath = file1.parent
    dirpath.mkdir()

    open(file1, "a").close()
    open(file2, "a").close()

    path = UPath("file://" + str(dirpath))

    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l2/", atype="dataset", data=dict(sensor="olci"))
    data_config = dict(unpack=dict(format="directory", include_dir_in_uid=True))

    with patched_publisher() as messages:
        file_publisher_from_generator([[path, dict()]],
                                      dict(publisher_config=publisher_settings,
                                           message_config=message_settings,
                                           data_config=data_config))

    assert "my_dir/file1" in messages[0]
    assert "my_dir/file2" in messages[0]
    msg = Message.decode(messages[0])
    assert msg.data["dataset"][0]["uid"].startswith("my_dir")


def test_publish_paths_with_fetching(tmp_path):
    """Test publishing paths."""
    basename = "foo+bar,baz_.txt"
    filename = os.fspath(tmp_path / basename)
    with open(filename, "w"):
        pass

    destination = tmp_path / "downloaded"
    destination.mkdir()

    publisher_settings = dict(nameservers=False, port=1979)
    subject = "/hey/jude"
    atype = "atomic"
    message_settings = dict(atype="file", data=dict(sensor="viirs"))
    data_config = dict(fetch=dict(destination=destination))

    items = [(filename, dict(subject=subject, atype=atype,
                             data=dict(mime="txt")))]

    with patched_publisher() as messages:
        config = dict(publisher_config=publisher_settings,
                      message_config=message_settings,
                      data_config=data_config)
        file_publisher_from_generator(items, config)

    assert "uri" not in message_settings["data"]
    assert len(messages) == 1
    message = Message(rawstr=messages[0])
    assert message.subject == subject
    assert message.type == "file"
    assert message.data["sensor"] == "viirs"



"""Tests for the fetching capability."""

import json
from urllib.parse import quote
from zipfile import ZipFile

import fsspec

from pytroll_watchers.fetch import fetch_file


def create_data_file(path):
    """Create a data file."""
    path.parent.mkdir()

    with open(path, "w") as fd:
        fd.write("data")


def test_fetcher_with_simple_uri(tmp_path):
    """Test fetcher with a simple uri."""
    filename = "important_data.nc"
    test_file = tmp_path / filename
    download_dir = tmp_path / "downloaded"
    download_dir.mkdir()
    with open(test_file, "w") as fd:
        fd.write("very important data.")
    uri = test_file.as_uri()
    fetch_file(uri, str(download_dir))
    assert (download_dir / filename).exists()


# def test_fetcher_with_decompression(tmp_path):
#     """Test fetcher with a simple uri."""
#     filename = "important_data.txt.bz2"
#     test_file = tmp_path / filename
#     download_dir = tmp_path / "downloaded"
#     download_dir.mkdir()
#     data = b"very important data."
#     with open(test_file, "wb") as fd:
#         fd.write(bz2.compress(data))
#     from zipfile import ZipFile
#     zfile = str(test_file) + ".zip"
#     with ZipFile(zfile, "w") as zipf:
#         zipf.write(test_file, arcname=filename)
#     of = fsspec.open("zip://" + filename + "::file://" + zfile)
#
#     fs = json.loads(of.fs.to_json())
#     fetch_file(of.path, download_dir, fs, compression="infer")
#     assert (download_dir / filename).open("rb").read() == data
#     # TODO
#     # assert str(download_dir / filename).endswith(".txt")
#
#
# def test_fetcher_with_different_dest_filename(tmp_path):
#     """Test fetcher with a simple uri."""
#     filename = "important_data.nc"
#     test_file = tmp_path / filename
#     download_dir = tmp_path / "downloaded"
#     download_dir.mkdir()
#     with open(test_file, "w") as fd:
#         fd.write("very important data.")
#     source_pattern = "{name}.nc"
#     destination_filename = "very_{name}.txt"
#     uri = test_file.as_uri()
#     new_name = construct_new_name(uri, source_pattern, destination_filename)
#     fetch_file(uri, new_name)
#     assert (download_dir / ("very_" + filename[:-2] + "txt")).exists()


def test_fetcher_with_complex_uri(tmp_path):
    """Test fetcher with a complex uri."""
    filename = "important_data.nc"
    test_file = tmp_path / filename
    download_dir = tmp_path / "downloaded"
    download_dir.mkdir()
    with open(test_file, "w") as fd:
        fd.write("very important data.")
    uri = "simplecache::" + test_file.as_uri()
    fetch_file(uri, download_dir)
    assert (download_dir / filename).exists()


def test_fetcher_with_annoying_filename(tmp_path):
    """Test fetcher with a filename containing symbols."""
    filename = "important+data,here.nc"
    test_file = tmp_path / filename
    download_dir = tmp_path / "downloaded"
    download_dir.mkdir()
    with open(quote(str(test_file)), "w") as fd:
        fd.write("very important data.")
    uri = test_file.as_uri()
    fetch_file(uri, str(download_dir))
    assert (download_dir / filename).exists()


def test_fetcher_with_fs(tmp_path):
    """Test fetcher with a filesystem."""
    filename = "important_data.nc"
    test_file = tmp_path / filename
    download_dir = tmp_path / "downloaded"
    download_dir.mkdir()
    with open(test_file, "w") as fd:
        fd.write("very important data.")
    uri = test_file.as_uri()
    fs = {"cls": "fsspec.implementations.local.LocalFileSystem", "protocol": "file", "args": []}
    fetch_file(uri, download_dir, fs)
    assert (download_dir / filename).exists()


def test_fetcher_with_complex_uri_and_fs(tmp_path):
    """Test fetcher with a complex uri and a filesystem."""
    filename = "important_data.nc"
    test_file = tmp_path / filename

    download_dir = tmp_path / "downloaded"
    download_dir.mkdir()

    with open(test_file, "w") as fd:
        fd.write("very important data.")

    compressed_test_file = tmp_path / "important.zip"
    with ZipFile(compressed_test_file, "w") as myzip:
        myzip.write(test_file)

    uri = "zip://" + str(test_file) + "::" + compressed_test_file.as_uri()
    fs = json.loads(fsspec.open(uri).fs.to_json())

    returned_filename = fetch_file(str(test_file), download_dir, fs)
    downloaded_filename = download_dir / filename
    assert downloaded_filename.exists()
    assert downloaded_filename == returned_filename


def test_fetcher_with_complex_uri_and_fs_2(tmp_path):
    """Test fetcher with a complex uri and a filesystem."""
    filename = "important_data.nc"
    test_file = tmp_path / filename

    download_dir = tmp_path / "downloaded"
    download_dir.mkdir()

    with open(test_file, "w") as fd:
        fd.write("very important data.")

    compressed_test_file = tmp_path / "important.zip"
    with ZipFile(compressed_test_file, "w") as myzip:
        myzip.write(test_file)

    uri = "zip://" + str(test_file) + "::" + compressed_test_file.as_uri()
    fs = json.loads(fsspec.open(uri).fs.to_json())

    returned_filename = fetch_file("zip://" + str(test_file), download_dir, fs)
    downloaded_filename = download_dir / filename
    assert downloaded_filename.exists()
    assert downloaded_filename == returned_filename



"""Tests for the selector."""
import time

import pytest
import yaml
from posttroll.message import Message
from posttroll.testing import patched_publisher, patched_subscriber_recv

from pytroll_watchers.selector import (
    TTLDict,
    _run_selector_with_managed_dict_server,
    _started_redis_server,
    cli,
    run_selector,
)


def test_ttldict_multiple_redis_instances(tmp_path):
    """Test the TTLDict."""
    ttl = 300
    key = "uid_multiple"
    value = b"some stuff"
    other_value = b"some other important stuff"
    port = 7321
    with _started_redis_server(port=port, directory=tmp_path / "redis_1"):
        sel = TTLDict(ttl, port=port)

        sel[key] = value
        assert sel[key] == value
        sel[key] = other_value
        assert sel[key] == value
    with _started_redis_server(port=port, directory=tmp_path / "redis_2"):
        with pytest.raises(KeyError):
            sel[key]


def test_redis_server_validates_directory(tmp_path):
    """Test the TTLDict."""
    port = 7321
    with _started_redis_server(port=port, directory=str(tmp_path / "redis_1")):
        assert True


def test_run_selector_that_starts_redis_on_given_port(tmp_path):
    """Test running a selector that also starts a redis server."""
    uid = "IVCDB_j03_d20240419_t1114110_e1115356_b07465_c20240419113435035578_cspp_dev.h5"
    sdr_file = tmp_path / "sdr" / uid
    create_data_file(sdr_file)

    msg1 = ('pytroll://segment/viirs/l1b/ file a001673@c22519.ad.smhi.se 2024-04-19T11:35:00.487388 v1.01 '
            'application/json {"sensor": "viirs", '
            f'"uid": "{uid}", "uri": "file://{str(sdr_file)}", "path": "{str(sdr_file)}", '
            '"filesystem": {"cls": "fsspec.implementations.local.LocalFileSystem", "protocol": "file", "args": []}}')

    messages = [Message.decode(msg1)]

    pipe_in_address = "ipc://" + str(tmp_path / "in.ipc")
    pipe_out_address = "ipc://" + str(tmp_path / "out.ipc")
    subscriber_config = dict(addresses=[pipe_in_address],
                             nameserver=False,
                             port=3000)

    publisher_config = dict(address=pipe_out_address,
                            nameservers=False,
                            port=1999)

    selector_config = dict(ttl=1, host="localhost", port=6388)

    with patched_subscriber_recv(messages):
        with patched_publisher() as published_messages:
            run_selector(selector_config, subscriber_config, publisher_config)
    assert len(published_messages) == 1


def test_run_selector_ignores_non_file_messages(tmp_path):
    """Test running a selector ignore irrelevant messages."""
    uid = "IVCDB_j04_d20240419_t1114110_e1115356_b07465_c20240419113435035578_cspp_dev.h5"
    sdr_file = tmp_path / "sdr" / uid
    create_data_file(sdr_file)

    msg1 = ('pytroll://segment/viirs/l1b/ del a001673@c22519.ad.smhi.se 2024-04-19T11:35:00.487388 v1.01 '
            'application/json {"sensor": "viirs", '
            f'"uid": "{uid}", "uri": "file://{str(sdr_file)}", "path": "{str(sdr_file)}", '
            '"filesystem": {"cls": "fsspec.implementations.local.LocalFileSystem", "protocol": "file", "args": []}}')

    messages = [Message.decode(msg1)]

    pipe_in_address = "ipc://" + str(tmp_path / "in.ipc")
    pipe_out_address = "ipc://" + str(tmp_path / "out.ipc")
    subscriber_config = dict(addresses=[pipe_in_address],
                             nameserver=False,
                             port=3000)

    publisher_config = dict(address=pipe_out_address,
                            nameservers=False,
                            port=2000)

    selector_config = dict(ttl=1, host="localhost", port=6298)

    with patched_subscriber_recv(messages):
        with patched_publisher() as published_messages:
            run_selector(selector_config, subscriber_config, publisher_config)
    assert len(published_messages) == 0


@pytest.fixture(scope="module")
def _redis_server():
    """Start a redis server."""
    with _started_redis_server():
        yield


def create_data_file(path):
    """Create a data file."""
    path.parent.mkdir(exist_ok=True)

    with open(path, "w") as fd:
        fd.write("data")


def test_run_selector_on_single_file_messages(tmp_path):
    """Test running the selector on single file messages."""
    uid = "IVCDB_j02_d20240419_t1114110_e1115356_b07465_c20240419113435035578_cspp_dev.h5"
    sdr_file = tmp_path / "sdr" / uid
    create_data_file(sdr_file)

    uid2 = "IVCDB_j01_d20240419_t1114110_e1115356_b07465_c20240419113435035578_cspp_dev.h5"
    sdr_file2 = tmp_path / "sdr" / uid2
    create_data_file(sdr_file2)


    msg1 = ('pytroll://segment/viirs/l1b/ file a001673@c22519.ad.smhi.se 2024-04-19T11:35:00.487388 v1.01 '
            'application/json {"sensor": "viirs", '
            f'"uid": "{uid}", "uri": "file://{str(sdr_file)}", "path": "{str(sdr_file)}", '
            '"filesystem": {"cls": "fsspec.implementations.local.LocalFileSystem", "protocol": "file", "args": []}}')

    msg2 = ('pytroll://segment/viirs/l1b/ file a001673@c22519.ad.smhi.se 2024-04-19T11:35:00.487388 v1.01 '
            'application/json {"sensor": "viirs", '
            f'"uid": "{uid}", "uri": "ssh://someplace.pytroll.org:/{str(sdr_file)}", "path": "{str(sdr_file)}", '
            '"filesystem": {"cls": "fsspec.implementations.ssh.SFTPFileSystem", "protocol": "ssh", "args": []}}')

    msg3 = ('pytroll://segment/viirs/l1b/ file a001673@c22519.ad.smhi.se 2024-04-19T11:35:00.487388 v1.01 '
            'application/json {"sensor": "viirs", '
            f'"uid": "{uid2}", "uri": "ssh://someplace.pytroll.org:/{str(sdr_file2)}", "path": "{str(sdr_file2)}", '
            '"filesystem": {"cls": "fsspec.implementations.ssh.SFTPFileSystem", "protocol": "ssh", "args": []}}')

    messages = [Message.decode(msg1), Message.decode(msg2), Message.decode(msg3)]

    pipe_in_address = "ipc://" + str(tmp_path / "in.ipc")
    pipe_out_address = "ipc://" + str(tmp_path / "out.ipc")
    subscriber_config = dict(addresses=[pipe_in_address],
                             nameserver=False,
                             port=3000)

    publisher_config = dict(address=pipe_out_address,
                            port=1999,
                            nameservers=False)

    selector_config = dict(ttl=1, host="localhost", port=6309)

    with patched_subscriber_recv(messages):
        with patched_publisher() as published_messages:
            with _started_redis_server(port=6309):
                _run_selector_with_managed_dict_server(selector_config, subscriber_config, publisher_config)
    assert len(published_messages) == 2
    assert published_messages[0] == msg1
    assert published_messages[1] == msg3


@pytest.mark.usefixtures("_redis_server")
def test_ttldict():
    """Test the TTLDict."""
    ttl = 1
    key = "uid_1"
    value = b"some stuff"
    other_value = b"some other important stuff"

    sel = TTLDict(ttl)

    sel[key] = value
    assert sel[key] == value
    sel[key] = other_value
    assert sel[key] == value
    time.sleep(ttl + 0.1)
    sel[key] = other_value
    assert sel[key] == other_value


def test_cli(tmp_path):
    """Test the command-line interface."""
    uid = "IVCDB_j03_d20240419_t1114110_e1115356_b07465_c20240419113435035578_cspp_dev.h5"
    sdr_file = tmp_path / "sdr" / uid
    create_data_file(sdr_file)

    msg1 = ('pytroll://segment/viirs/l1b/ file a001673@c22519.ad.smhi.se 2024-04-19T11:35:00.487388 v1.01 '
            'application/json {"sensor": "viirs", '
            f'"uid": "{uid}", "uri": "file://{str(sdr_file)}", "path": "{str(sdr_file)}", '
            '"filesystem": {"cls": "fsspec.implementations.local.LocalFileSystem", "protocol": "file", "args": []}}')

    messages = [Message.decode(msg1)]

    pipe_in_address = "ipc://" + str(tmp_path / "in.ipc")
    pipe_out_address = "ipc://" + str(tmp_path / "out.ipc")
    subscriber_config = dict(addresses=[pipe_in_address],
                             nameserver=False,
                             port=3000)

    publisher_config = dict(address=pipe_out_address,
                            nameservers=False,
                            port=1999)

    redis_dir = tmp_path / "redis_dir"

    selector_config = dict(ttl=1, port=6389, directory=str(redis_dir))
    config = dict(publisher_config=publisher_config,
                  subscriber_config=subscriber_config,
                  selector_config=selector_config)
    config_file = tmp_path / "selector_config"
    with open(config_file, "w") as fd:
        fd.write(yaml.dump(config))
    with patched_subscriber_recv(messages):
        with patched_publisher() as published_messages:
            cli([str(config_file)])
    assert len(published_messages) == 1



"""Tests for the dhus watcher."""


import datetime as dt

import responses
import responses._recorder
from freezegun import freeze_time
from posttroll.message import Message
from posttroll.testing import patched_publisher

from pytroll_watchers.dhus_watcher import (
    file_generator,
    file_publisher,
    generate_download_links,
    generate_download_links_since,
)

server = "https://colhub.met.no"


@responses.activate
def test_generate_download_links():
    """Test the generation of download links."""
    responses._add_from_file(file_path="tests/dhus_responses.yaml")

    filter_params = ["substringof('IW_GRDH',Name)",
                     "IngestionDate gt datetime'2024-04-29T07:33:00.000'"]

    links = list(generate_download_links(server, filter_params))
    assert len(links) == 5
    path, mda = links[0]
    assert path.as_uri() == "https://colhub.met.no/odata/v1/Products('e49f8de5-1647-4a7b-ba69-268f9aa77f42')/$value"
    assert mda["boundary"]["coordinates"][0][0] == [-4.545108, 57.399067]

    assert mda["platform_name"] == "Sentinel-1A"
    assert mda["sensor"] == "SAR"
    assert "ingestion_date" in mda
    assert mda["product_type"] == "IW_GRDH_1S"
    assert mda["start_time"] == dt.datetime(2024, 4, 29, 6, 46, 23, 104000, tzinfo=dt.timezone.utc)
    assert mda["end_time"] == dt.datetime(2024, 4, 29, 6, 46, 48, 102000, tzinfo=dt.timezone.utc)
    assert mda["orbit_number"] == 53645
    assert "checksum" in mda
    assert "size" in mda


@responses.activate
def test_generate_download_links_since():
    """Test limiting the download links to a given point in time."""
    responses._add_from_file(file_path="tests/dhus_responses.yaml")

    filter_params = ["substringof('IW_GRDH',Name)"]
    since = dt.datetime(2024, 4, 29, 7, 33, tzinfo=dt.timezone.utc)
    links = list(generate_download_links_since(server, filter_params, since))
    assert len(links) == 5
    path, mda = links[1]
    assert path.as_uri() == "https://colhub.met.no/odata/v1/Products('162f165a-23c3-4212-869e-4f38b9dade63')/$value"
    assert mda["start_time"] == dt.datetime(2024, 4, 29, 6, 52, 38, 105000, tzinfo=dt.timezone.utc)


@freeze_time(dt.datetime(2024, 4, 29, 8, 33, tzinfo=dt.timezone.utc))
@responses.activate
def test_file_generator():
    """Test the file generator."""
    responses._add_from_file(file_path="tests/dhus_responses.yaml")

    filter_params = ["substringof('IW_GRDH',Name)"]
    since = dt.timedelta(hours=1)
    polling_interval = dict(minutes=0)
    links = list(file_generator(server, filter_params, polling_interval, since))
    assert len(links) == 5
    path, mda = links[3]
    assert path.as_uri() == "https://colhub.met.no/odata/v1/Products('a6b214dd-fe39-4f50-aa94-f84d8510cedc')/$value"
    assert mda["start_time"] == dt.datetime(2024, 4, 29, 6, 54, 18, 105000, tzinfo=dt.timezone.utc)
    assert "ingestion_date" not in mda
    assert mda["uid"] == "S1A_IW_GRDH_1SDV_20240429T065418_20240429T065443_053645_0683C0_8D1D.SAFE"


@freeze_time(dt.datetime(2024, 4, 29, 8, 33, tzinfo=dt.timezone.utc))
@responses.activate
def test_publish_paths():
    """Test publishing paths."""
    responses._add_from_file(file_path="tests/dhus_responses.yaml")

    filter_params = ["substringof('IW_GRDH',Name)"]

    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l1b/", atype="file", aliases=dict(sensor={"SAR": "SAR-C"}))

    with patched_publisher() as messages:
        fs_config = dict(server=server,
                         filter_params=filter_params,
                         polling_interval=dict(minutes=0),
                         start_from=dict(hours=1))


        file_publisher(dict(fs_config=fs_config,
                            publisher_config=publisher_settings,
                            message_config=message_settings))
    message = Message(rawstr=messages[3])

    assert message.data["uid"] == "S1A_IW_GRDH_1SDV_20240429T065418_20240429T065443_053645_0683C0_8D1D.SAFE"
    assert message.data["sensor"] == "SAR-C"
